""":class:`INIParser` ini parser implementation.
"""
from __future__ import annotations
import os
from configparser import ConfigParser
import re
from typing import Any, Dict, ItemsView, KeysView, List, TypeVar, Type, Union, ValuesView
import numpy as np

T = TypeVar('T')            # Object type
Desc = TypeVar('Desc')      # Descriptor type
ROOT_PATH = os.path.dirname(__file__)

class hybridmethod:
    """Hybrid method descriptor supporting
    two distinct methodsbound to class and instance.

    Attributes:
        fclass : Class bound method.
        finstance : Instance bound method.
        doc : Method's dosctring.
    """
    def __init__(self, fclass: Desc, finstance: Desc=None, doc: str=None) -> None:
        """Args:
            fclass : Class bound method.
            finstance : Instance bound method.
            doc : Method's docstring.
        """
        self.fclass, self.finstance = fclass, finstance
        self.__doc__ = doc or fclass.__doc__
        self.__isabstractmethod__ = bool(getattr(fclass, '__isabstractmethod__', False))

    def classmethod(self, fclass: Desc) -> hybridmethod:
        """Class method decorator

        Args:
            fclass : Class bound method.

        Returns:
            A new instance with the class bound method added
            to the object.
        """
        return type(self)(fclass, self.finstance, None)

    def instancemethod(self, finstance: Desc) -> hybridmethod:
        """Instance method decorator

        Args:
            finstance : Instance bound method.

        Returns:
            A new instance with the instance bound method added
            to the object.
        """
        return type(self)(self.fclass, finstance, self.__doc__)

    def __get__(self, instance: T, cls: Type[T]) -> Any:
        if instance is None or self.finstance is None:
            return self.fclass.__get__(cls, None)
        return self.finstance.__get__(instance, cls)

class INIParser:
    """INI files parser class with the methods for importing and exporting
    ini files and Python dictionaries.

    Attributes:
        err_txt : Error text.
        known_types : Look-up dictionary of supported types for formatting.
        attr_dict : Dictionary of provided attributes.
        fmt_dict : Dictionary of attributes' types used for formatting.
    """
    err_txt = "Wrong format key '{0:s}' of option '{1:s}'"
    known_types = {'int': int, 'bool': bool, 'float': float, 'str': str}
    attr_dict, fmt_dict = {}, {}
    LIST_SPLITTER = r'\s*,\s*'
    LIST_MATCHER = r'^\[([\s\S]*)\]$'
    FMT_LEN = 3

    def __init__(self, **kwargs: Any) -> None:
        """Args:
            kwargs : Attributes specified in `attr_dict`.

        Raises:
            AttributeError : If the attribute specified in `attr_dict`
                has not been provided in keyword arguments ``**kwargs``.
        """
        self.__dict__['_lookup'] = {}
        self.__dict__['ini_dict'] = {section: {} for section in self.attr_dict}
        for section in self.attr_dict:
            if 'ALL' in self.attr_dict[section] and section in kwargs:
                value = {option: self._get_value(section, option, kwargs)
                         for option in kwargs[section]}
            else:
                value = {option: self._get_value(section, option, kwargs)
                         for option in self.attr_dict[section]}
            self.__setattr__(section, value)
        self._lookup = self._lookup_dict()

    def _get_value(self, section: str, option: str, kwargs: Dict) -> Any:
        if not option in kwargs[section]:
            raise AttributeError("The '{:s}' option has not been provided".format(option))
        fmt = self.get_format(section, option)

        if isinstance(kwargs[section][option], (list, tuple)):
            return [fmt(part) for part in kwargs[section][option]]

        if isinstance(kwargs[section][option], np.ndarray):
            if kwargs[section][option].ndim > 1:
                raise ValueError(f'{kwargs[section][option]:s} must be one-dimensional')
            return [fmt(part) for part in kwargs[section][option]]

        return fmt(kwargs[section][option])

    @staticmethod
    def str_to_list(strings: Union[str, List[str]]) -> List[str]:
        """Convert `strings` to a list of strings.

        Args:
            strings : String or a list of strings

        Returns:
            List of strings.
        """
        if isinstance(strings, (str, list)):
            if isinstance(strings, str):
                return [strings,]
            return strings

        raise ValueError('strings must be a string or a list of strings')

    @classmethod
    def _lookup_dict(cls) -> Dict:
        """Look-up table between the sections and the parameters.

        Returns:
            Look-up dictionary.
        """
        return {}

    @classmethod
    def read_ini(cls, protocol_file: str) -> ConfigParser:
        """Read the `protocol_file` and return an instance of
        :class:`configparser.ConfigParser` class.

        Args:
            protocol_file : Path to the file.

        Returns:
            Parser object with all the data contained in the
            INI file.

        Raises:
            ValueError : If the file doesn't exist.
        """
        if not os.path.isfile(protocol_file):
            raise ValueError(f"File {protocol_file} doesn't exist")
        ini_parser = ConfigParser()
        ini_parser.read(protocol_file)
        return ini_parser

    @classmethod
    def get_format(cls, section: str, option: str) -> Type:
        """Return the attribute's format specified by `fmt_dict`.

        Args:
            section : Attribute's section.
            option : Attribute's name.

        Returns:
            Type of the attribute.
        """
        fmt = cls.fmt_dict.get(os.path.join(section, option))
        if not fmt:
            fmt = cls.fmt_dict.get(section)
        return cls.known_types.get(fmt, str)

    @classmethod
    def get_value(cls, ini_parser: ConfigParser, section: str, option: str) -> Any:
        """Return an attribute from an INI file's parser object `ini_parser`.

        Args:
            ini_parser : A parser object of an INI file.
            section : Attribute's section.
            option : Attribute's option.

        Returns:
            Attribute's value imported from the INI file.
        """
        fmt = cls.get_format(section, option)
        string = ini_parser.get(section, option)
        is_list = re.search(cls.LIST_MATCHER, string)
        if is_list:
            return [fmt(part.strip('\'\"'))
                    for part in re.split(cls.LIST_SPLITTER, is_list.group(1))]

        return fmt(string.strip())

    @classmethod
    def _import_ini(cls, protocol_file: str) -> Dict[str, Dict]:
        ini_parser = cls.read_ini(protocol_file)
        kwargs = {}
        for section in cls.attr_dict:
            kwargs[section] = {}
            if 'ALL' in cls.attr_dict[section]:
                if section in ini_parser:
                    for option in ini_parser[section]:
                        val = cls.get_value(ini_parser, section, option)
                        kwargs[section][option] = val
            else:
                for option in cls.attr_dict[section]:
                    val = cls.get_value(ini_parser, section, option)
                    kwargs[section][option] = val
        return kwargs

    @classmethod
    def _format(cls, obj: Dict) -> Dict:
        crop_obj = {}
        for key, val in list(obj.items())[:cls.FMT_LEN]:
            if isinstance(val, dict):
                val = cls._format(val)
            elif isinstance(val, list) and len(val) > cls.FMT_LEN:
                val = val[:cls.FMT_LEN] + ['...']
            crop_obj[key] = val
        if len(obj) > cls.FMT_LEN:
            crop_obj['...'] = '...'
        return crop_obj

    def __getattr__(self, attr: str) -> Any:
        if attr in self.__dict__.get('_lookup', {}):
            return self.ini_dict[self.__dict__['_lookup'][attr]][attr]
        if attr in self.__dict__.get('ini_dict', {}):
            return self.__dict__['ini_dict'][attr]
        raise AttributeError(attr + " doesn't exist")

    def __getitem__(self, attr: str) -> Any:
        return self.__getattr__(attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in self._lookup:
            self.ini_dict[self._lookup[attr]][attr] = value
        elif attr in self.ini_dict:
            self.ini_dict[attr] = value
        else:
            super(INIParser, self).__setattr__(attr, value)

    def __repr__(self) -> str:
        crop_dict = {key: self._format(val) for key, val in self.export_dict().items()}
        return crop_dict.__repr__()

    def __str__(self) -> str:
        crop_dict = {key: self._format(val) for key, val in self.export_dict().items()}
        return crop_dict.__str__()

    def keys(self) -> KeysView:
        return self.attr_dict.keys()

    def items(self) -> ItemsView:
        """Return (key, value) pairs of the datasets stored in the container.

        Returns:
            (key, value) pairs of the datasets stored in the container.
        """
        return dict(self).items()

    def values(self) -> ValuesView:
        """Return the attributes' data stored in the container.

        Returns:
            List of data stored in the container.
        """
        return dict(self).values()

    def export_dict(self) -> Dict[str, Any]:
        """Return a :class:`dict` object with all the attributes.

        Returns:
            Dictionary with all the attributes conntained in the object.
        """
        return dict(self)

    @hybridmethod
    def export_ini(self, **kwargs: Dict) -> ConfigParser:
        """Return a :class:`configparser.ConfigParser` object
        with all the attributes exported from the class.

        Args:
            kwargs : Extra parameters to export to the
                :class:`configparser.ConfigParser` object.

        Returns:
            A parser object with all the parsing specifications
            contained in class.
        """
        ini_parser = ConfigParser()
        for section in self.attr_dict:
            if 'ALL' in self.attr_dict[section]:
                ini_parser[section] = kwargs[section]
            else:
                ini_parser[section] = {option: kwargs[section][option]
                                       for option in self.attr_dict[section]}
        return ini_parser

    @export_ini.instancemethod
    def export_ini(self) -> ConfigParser:
        """Return a :class:`configparser.ConfigParser` object
        with all the attributes exported from the object.

        Returns:
            A parser object with all the parsing specifications
            contained in the object.
        """
        return type(self).export_ini(**self.ini_dict)



""":class:`DataContainer` class implementation.
"""
from __future__ import annotations
from typing import (Any, Callable, Dict, ItemsView, Iterable, Iterator,
                    List, Optional, ValuesView, TypeVar, Type)

T = TypeVar('T')

class BoundMethod:
    __func__: Callable[..., Dict]
    __self__: T

class dict_to_object:
    """Creates a new bound method. Wraps a function implementation of a
    class bound method to return an instance of an object instead of a
    dictionary.

    Attributes:
        finstance : Class bound method.
    """
    def __init__(self, finstance: Callable[..., Dict]) -> None:
        """
        Args:
            finstance : Function object containing implementation of the
                class bound method.
        """
        self.finstance = finstance

    def __get__(self, instance: T, cls: Type[T]):
        return return_obj_method(self.finstance.__get__(instance, cls), instance, cls)

class return_obj_method:
    """Factory class that uses a bound method to return an instance of an
    object instead of a dictionary.

    Attributes:
        instance : Object instance.
        cls : Object class.
        sig_attrs : Signature attributes.
    """
    sig_attrs = {'__annotations__', '__doc__', '__module__',
                 '__name__', '__qualname__'}

    def __init__(self, method: BoundMethod, instance: T, cls: Type[T]) -> None:
        """
        Args:
            method : Wrapped method that returns a dictionary.
            instance : Object instance.
            cls : Object class.
        """
        self.instance, self.cls = instance, cls
        self.__wrapped__ = method
        for attr in self.sig_attrs:
            self.__dict__[attr] = getattr(method, attr)

    def __call__(self, *args: Any, **kwargs: Any) -> T:
        """Return an object from the dictionary yielded by
        the wrapped method.

        Args:
            args : Positional arguments.
            kwargs : Keyword arguments.

        Returns:
            A new object instance.
        """
        dct = {}
        dct.update(self.__wrapped__(*args, **kwargs))
        for key, val in self.instance.items():
            if key not in dct:
                dct[key] = val
        return self.cls(**dct)

    def inplace_update(self, *args: Any, **kwargs: Any) -> None:
        """Modify the object by the dictionary yielded from
        the wrapped method.

        Args:
            args : Positional arguments.
            kwargs : Keyword arguments.
        """
        dct = self.__wrapped__(*args, **kwargs)
        for key, val in dct.items():
            self.instance.__setattr__(key, val)

class DataContainer:
    """Abstract data container class.

    Attributes:
        attr_set : Set of attributes in the container which are necessary
            to initialize in the constructor.
        init_set : Set of optional data attributes.
    """
    attr_set, init_set = set(), set()

    def __init__(self, **kwargs: Any) -> None:
        """
        Args:
            kwargs : Values of the attributes specified in `attr_set` and
                `init_set`.

        Raises:
            ValueError : If an attribute specified in `attr_set` has not been
                provided.
        """
        for attr in self.attr_set:
            if kwargs.get(attr, None) is None:
                raise ValueError(f'Attribute {attr} has not been provided')

        for attr in self.init_set:
            self.__dict__[attr] = None

        for attr in kwargs:
            if attr in self:
                self.__setattr__(attr, kwargs.get(attr))
            else:
                raise ValueError(f'Parameter {attr} is invalid')

        self.init_funcs = {}

    def _init_functions(self, **kwargs: Callable) -> None:
        self.init_funcs.update(**kwargs)

    def _init_attributes(self) -> None:
        for attr, init_func in self.init_funcs.items():
            if self.__dict__.get(attr, None) is None:
                self.__setattr__(attr, init_func())

    def __iter__(self) -> Iterator[str]:
        return (self.attr_set | self.init_set).__iter__()

    def __contains__(self, attr: str) -> bool:
        return attr in self.attr_set | self.init_set

    def __getitem__(self, attr: str) -> Any:
        return self.__dict__.__getitem__(attr)

    def __repr__(self) -> str:
        return {attr: self.__dict__[attr] for attr in self
                if self.__dict__.get(attr) is not None}.__repr__()

    def __str__(self) -> str:
        return {attr: self.__dict__[attr] for attr in self
                if self.__dict__.get(attr) is not None}.__str__()

    def get(self, attr: str, value: Optional[Any]=None) -> Any:
        """Retrieve a dataset, return `value` if the attribute is not found.

        Args:
            attr : Data attribute.
            value : Data which is returned if the attribute is not found.

        Returns:
            Attribute's data stored in the container, `value` if `attr`
            is not found.
        """
        return self.__dict__.get(attr, value)

    def contents(self) -> List[str]:
        """Return a list of the attributes stored in the container.

        Returns:
            List of the attributes stored in the container.
        """
        return [attr for attr in self if self.get(attr) is not None]

    def keys(self) -> Iterable[str]:
        """Return a list of the attributes available in the container.

        Returns:
            List of the attributes available in the container.
        """
        return list(self)

    def items(self) -> ItemsView:
        """Return (key, value) pairs of the datasets stored in the container.

        Returns:
            (key, value) pairs of the datasets stored in the container.
        """
        return dict(self).items()

    def values(self) -> ValuesView:
        """Return the attributes' data stored in the container.

        Returns:
            List of data stored in the container.
        """
        return dict(self).values()



# Copyright 2014 Knowledge Economy Developments Ltd
#
# Henry Gomersall
# heng@kedevelopments.co.uk
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#

cimport numpy as np
from libc.stdint cimport int64_t

ctypedef struct _fftw_iodim:
    int _n
    int _is
    int _os

ctypedef float cfloat[2]
ctypedef double cdouble[2]

cdef extern from "cpu.h":

    int simd_alignment()

cdef extern from "Python.h":
    int Py_AtExit(void(*func)())

cdef extern from 'fftw3.h':

    # Double precision plans
    ctypedef struct fftw_plan_struct:
        pass

    ctypedef fftw_plan_struct *fftw_plan

    # Single precision plans
    ctypedef struct fftwf_plan_struct:
        pass

    ctypedef fftwf_plan_struct *fftwf_plan

    # The stride info structure. I think that strictly
    # speaking, this should be defined with a type suffix
    # on fftw (ie fftw, fftwf or fftwl), but since the
    # definition is transparent and is defined as _fftw_iodim,
    # we ignore the distinction in order to simplify the code.
    ctypedef struct fftw_iodim:
        pass

    # Double precision complex planner
    fftw_plan fftw_plan_guru_dft(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            cdouble *_in, cdouble *_out,
            int sign, unsigned flags) nogil

    # Single precision complex planner
    fftwf_plan fftwf_plan_guru_dft(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            cfloat *_in, cfloat *_out,
            int sign, unsigned flags) nogil

    # Double precision real to complex planner
    fftw_plan fftw_plan_guru_dft_r2c(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            double *_in, cdouble *_out,
            unsigned flags) nogil

    # Single precision real to complex planner
    fftwf_plan fftwf_plan_guru_dft_r2c(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            float *_in, cfloat *_out,
            unsigned flags) nogil

    # Double precision complex to real planner
    fftw_plan fftw_plan_guru_dft_c2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            cdouble *_in, double *_out,
            unsigned flags) nogil

    # Single precision complex to real planner
    fftwf_plan fftwf_plan_guru_dft_c2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            cfloat *_in, float *_out,
            unsigned flags) nogil

    # Double precision real planner
    fftw_plan fftw_plan_guru_r2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            double *_in, double *_out,
            int *kind, unsigned flags)

    # Single precision real planner
    fftwf_plan fftwf_plan_guru_r2r(
            int rank, fftw_iodim *dims,
            int howmany_rank, fftw_iodim *howmany_dims,
            float *_in, float *_out,
            int *kind, unsigned flags)
    # Double precision complex new array execute
    void fftw_execute_dft(fftw_plan,
          cdouble *_in, cdouble *_out) nogil

    # Single precision complex new array execute
    void fftwf_execute_dft(fftwf_plan,
          cfloat *_in, cfloat *_out) nogil

    # Double precision real to complex new array execute
    void fftw_execute_dft_r2c(fftw_plan,
          double *_in, cdouble *_out) nogil

    # Single precision real to complex new array execute
    void fftwf_execute_dft_r2c(fftwf_plan,
          float *_in, cfloat *_out) nogil

    # Double precision complex to real new array execute
    void fftw_execute_dft_c2r(fftw_plan,
          cdouble *_in, double *_out) nogil

    # Single precision complex to real new array execute
    void fftwf_execute_dft_c2r(fftwf_plan,
          cfloat *_in, float *_out) nogil

    # Double precision real new array execute
    void fftw_execute_r2r(fftw_plan,
          double *_in, double *_out) nogil

    # Single precision real new array execute
    void fftwf_execute_r2r(fftwf_plan,
          float *_in, float *_out) nogil

    # Double precision plan destroyer
    void fftw_destroy_plan(fftw_plan)

    # Single precision plan destroyer
    void fftwf_destroy_plan(fftwf_plan)

    # Double precision set timelimit
    void fftw_set_timelimit(double seconds)

    # Single precision set timelimit
    void fftwf_set_timelimit(double seconds)

    # Threading routines
    # Double precision
    void fftw_init_threads()
    void fftw_plan_with_nthreads(int n)

    # Single precision
    void fftwf_init_threads()
    void fftwf_plan_with_nthreads(int n)

    # cleanup routines
    void fftw_cleanup()
    void fftwf_cleanup()
    void fftw_cleanup_threads()
    void fftwf_cleanup_threads()

    double FFTW_NO_TIMELIMIT

# Define function pointers that can act as a placeholder
# for whichever dtype is used (the problem being that fftw
# has different function names and signatures for all the
# different precisions and dft types).
ctypedef void * (*fftw_generic_plan_guru)(
        int rank, fftw_iodim *dims,
        int howmany_rank, fftw_iodim *howmany_dims,
        void *_in, void *_out,
        int *directions, unsigned flags) nogil

ctypedef void (*fftw_generic_execute)(void *_plan, void *_in, void *_out) nogil

ctypedef void (*fftw_generic_destroy_plan)(void *_plan)

ctypedef void (*fftw_generic_init_threads)()

ctypedef void (*fftw_generic_plan_with_nthreads)(int n)

ctypedef void (*fftw_generic_set_timelimit)(double seconds)

ctypedef bint (*validator)(np.ndarray input_array,
        np.ndarray output_array, int64_t *axes, int64_t *not_axes,
        int64_t axes_length)

# Direction enum
cdef enum:
    FFTW_FORWARD = -1
    FFTW_BACKWARD = 1
    # from fftw3.f 3.3.3; may not be valid for different versions of FFTW.
    FFTW_REDFT00  = 3
    FFTW_REDFT01  = 4
    FFTW_REDFT10  = 5
    FFTW_REDFT11  = 6
    FFTW_RODFT00  = 7
    FFTW_RODFT01  = 8
    FFTW_RODFT10  = 9
    FFTW_RODFT11  = 10

# Documented flags
cdef enum:
    FFTW_MEASURE = 0
    FFTW_DESTROY_INPUT = 1
    FFTW_UNALIGNED = 2
    FFTW_CONSERVE_MEMORY = 4
    FFTW_EXHAUSTIVE = 8
    FFTW_PRESERVE_INPUT = 16
    FFTW_PATIENT = 32
    FFTW_ESTIMATE = 64
    FFTW_WISDOM_ONLY = 2097152

cdef class FFTW:
    # Each of these function pointers simply
    # points to a chosen fftw wrapper function
    cdef fftw_generic_plan_guru _fftw_planner
    cdef fftw_generic_execute _fftw_execute
    cdef fftw_generic_destroy_plan _fftw_destroy
    cdef fftw_generic_plan_with_nthreads _nthreads_plan_setter

    # The plan is typecast when it is created or used
    # within the wrapper functions
    cdef void *_plan

    cdef np.ndarray _input_array
    cdef np.ndarray _output_array
    cdef int *_direction
    cdef unsigned _flags

    cdef bint _simd_allowed
    cdef int _input_array_alignment
    cdef int _output_array_alignment
    cdef bint _use_threads

    cdef object _input_item_strides
    cdef object _input_strides
    cdef object _output_item_strides
    cdef object _output_strides
    cdef object _input_shape
    cdef object _output_shape
    cdef object _input_dtype
    cdef object _output_dtype
    cdef object _flags_used

    cdef double _normalisation_scaling
    cdef double _sqrt_normalisation_scaling

    cdef int _rank
    cdef _fftw_iodim *_dims
    cdef int _howmany_rank
    cdef _fftw_iodim *_howmany_dims

    cdef int64_t *_axes
    cdef int64_t *_not_axes

    cdef int64_t _total_size

    cdef bint _normalise_idft
    cdef bint _ortho

    cdef inline _update_arrays(self, np.ndarray new_input_array,
                               np.ndarray new_output_array):
        ''' A C interface to the update_arrays method that does not
        perform any checks on strides being correct and so on.
        '''
        self._input_array = new_input_array
        self._output_array = new_output_array

    cdef inline _execute(self):
        '''execute()

        Execute the planned operation, taking the correct kind of FFT of
        the input array (i.e. :attr:`FFTW.input_array`),
        and putting the result in the output array (i.e.
        :attr:`FFTW.output_array`).
        '''
        cdef void *input_pointer = (
                <void *>np.PyArray_DATA(self._input_array))
        cdef void *output_pointer = (
                <void *>np.PyArray_DATA(self._output_array))

        cdef void *plan = self._plan
        cdef fftw_generic_execute fftw_execute = self._fftw_execute
        with nogil:
            fftw_execute(plan, input_pointer, output_pointer)


"""CXI protocol (:class:`pyrost.CXIProtocol`) is a helper class for a :class:`pyrost.STData`
data container, which tells it where to look for the necessary data fields in a CXI
file. The class is fully customizable so you can tailor it to your particular data
structure of CXI file.

Examples:
    Generate the default built-in CXI protocol as follows:

    >>> import pyrost as rst
    >>> rst.CXIProtocol.import_default()
    {'config': {'float_precision': 'float64'}, 'datatypes': {'basis_vectors': 'float',
    'data': 'uint', 'defocus_x': 'float', '...': '...'}, 'default_paths': {'basis_vectors':
    '/speckle_tracking/basis_vectors', 'data': '/entry/data/data', 'defocus_x':
    '/speckle_tracking/defocus_x', '...': '...'}, 'is_data': {'basis_vectors': 'False',
    'data': 'True', 'defocus_x': 'False', '...': '...'}}
"""
from __future__ import annotations
from functools import partial
from multiprocessing import Pool
import os
from configparser import ConfigParser
from types import TracebackType
from typing import (Callable, Dict, ItemsView, Iterable, Iterator, KeysView,
                    List, Optional, Tuple, Union, ValuesView)
import h5py
import numpy as np
from tqdm.auto import tqdm
from .ini_parser import ROOT_PATH, INIParser

CXI_PROTOCOL = os.path.join(ROOT_PATH, 'config/cxi_protocol.ini')
Indices = Union[int, slice, np.ndarray, List[int], Tuple[int]]

class CXIProtocol(INIParser):
    """CXI protocol class. Contains a CXI file tree path with
    the paths written to all the data attributes necessary for
    the Speckle Tracking algorithm, corresponding attributes'
    data types, and floating point precision.

    Attributes:
        config : Protocol configuration.
        datatypes : Dictionary with attributes' datatypes. 'float', 'int', 'uint',
            or 'bool' are allowed.
        is_data : Dictionary with the flags if the attribute is of data type.
            Data type is 2- or 3-dimensional and has the same data shape
            as `data`.
        load_paths : Dictionary with attributes' CXI default file paths.
    """
    known_types = {'int': np.integer, 'bool': bool, 'float': np.floating, 'str': str,
                   'uint': np.unsignedinteger}
    default_types = {'int': np.int64, 'bool': bool, 'float': np.float64, 'str': str,
                     'uint': np.uint64}
    known_ndims = {'stack': (3,), 'frame': (2, 3), 'sequence': (1, 2, 3), 'scalar': (0, 1, 2)}
    attr_dict = {'datatypes': ('ALL', ), 'load_paths': ('ALL', ), 'kinds': ('ALL', )}
    fmt_dict = {'datatypes': 'str', 'load_paths': 'str', 'kinds': 'str'}

    datatypes   : Dict[str, str]
    load_paths  : Dict[str, List[str]]
    kinds       : Dict[str, str]

    def __init__(self, datatypes: Dict[str, str], load_paths: Dict[str, Union[str, List[str]]],
                 kinds: Dict[str, str]) -> None:
        """
        Args:
            datatypes : Dictionary with attributes' datatypes. 'float', 'int', 'uint',
                or 'bool' are allowed.
            load_paths : Dictionary with attributes path in the CXI file.
            kinds : Dictionary with the flags if the attribute is of data type.
                Data type is 2- or 3-dimensional and has the same data shape
                as `data`.
        """
        load_paths = {attr: self.str_to_list(val)
                      for attr, val in load_paths.items() if attr in datatypes}
        kinds = {attr: val for attr, val in kinds.items() if attr in datatypes}
        super(CXIProtocol, self).__init__(datatypes=datatypes, load_paths=load_paths,
                                          kinds=kinds)

    @classmethod
    def import_default(cls, datatypes: Optional[Dict[str, str]]=None,
                       load_paths: Optional[Dict[str, Union[str, List[str]]]]=None,
                       kinds: Optional[Dict[str, str]]=None) -> CXIProtocol:
        """Return the default :class:`CXIProtocol` object. Extra arguments
        override the default values if provided.

        Args:
            datatypes : Dictionary with attributes' datatypes. 'float', 'int', 'uint',
                or 'bool' are allowed.
            load_paths : Dictionary with attributes path in the CXI file.
            kinds : Dictionary with the flags if the attribute is of data type.
                Data type is 2- or 3-dimensional and has the same data shape
                as `data`.

        Returns:
            A :class:`CXIProtocol` object with the default parameters.
        """
        return cls.import_ini(CXI_PROTOCOL, datatypes, load_paths, kinds)

    @classmethod
    def import_ini(cls, ini_file: str, datatypes: Optional[Dict[str, str]]=None,
                   load_paths: Optional[Dict[str, Union[str, List[str]]]]=None,
                   kinds: Optional[Dict[str, str]]=None) -> CXIProtocol:
        """Initialize a :class:`CXIProtocol` object class with an
        ini file.

        Args:
            ini_file : Path to the ini file. Load the default CXI protocol if None.
            datatypes : Dictionary with attributes' datatypes. 'float', 'int', 'uint',
                or 'bool' are allowed. Initialized with `ini_file` if None.
            load_paths : Dictionary with attributes path in the CXI file. Initialized
                with `ini_file` if None.
            kinds : Dictionary with the flags if the attribute is of data type.
                Data type is 2- or 3-dimensional and has the same data shape
                as `data`. Initialized with `ini_file` if None.

        Returns:
            A :class:`CXIProtocol` object with all the attributes imported
            from the ini file.
        """
        kwargs = cls._import_ini(ini_file)
        if not datatypes is None:
            kwargs['datatypes'].update(**datatypes)
        if not load_paths is None:
            kwargs['load_paths'].update(**load_paths)
        if not kinds is None:
            kwargs['kinds'].update(**kinds)
        return cls(datatypes=kwargs['datatypes'], load_paths=kwargs['load_paths'],
                   kinds=kwargs['kinds'])

    def find_path(self, attr: str, cxi_file: h5py.File) -> str:
        """Find attribute's path in a CXI file `cxi_file`.

        Args:
            attr : Data attribute.
            cxi_file : :class:`h5py.File` object of the CXI file.

        Returns:
            Atrribute's path in the CXI file, returns an empty
            string if the attribute is not found.
        """
        paths = self.get_load_paths(attr, list())
        for path in paths:
            if path in cxi_file:
                return path
        return str()

    def parser_from_template(self, path: str) -> ConfigParser:
        """Return a :class:`configparser.ConfigParser` object using
        an ini file template.

        Args:
            path : Path to the ini file template.

        Returns:
            Parser object with the attributes populated according
            to the protocol.
        """
        ini_template = ConfigParser()
        ini_template.read(path)
        parser = ConfigParser()
        for section in ini_template:
            parser[section] = {option: ini_template[section][option].format(**self.default_paths)
                               for option in ini_template[section]}
        return parser

    def __iter__(self) -> Iterator:
        return self.datatypes.__iter__()

    def __contains__(self, attr: str) -> bool:
        return attr in self.datatypes

    def get_load_paths(self, attr: str, value: Optional[List[str]]=None) -> List[str]:
        """Return the atrribute's default path in the CXI file.
        Return `value` if `attr` is not found.

        Args:
            attr : The attribute to look for.
            value : Value which is returned if the `attr` is not found.

        Returns:
            Attribute's cxi file path.
        """
        return self.load_paths.get(attr, value)

    def get_dtype(self, attr: str, dtype: Optional[np.dtype]=None) -> type:
        """Return the attribute's data type. Return `dtype` if the attribute's
        data type is not found.

        Args:
            attr : The data attribute.
            dtype : Data type which is returned if the attribute's data type
                is not found.

        Returns:
            Attribute's data type.
        """
        if attr not in self:
            raise ValueError(f'Invalid attribute: {attr:s}')

        attr_type = self.known_types.get(self.datatypes.get(attr))
        if dtype is not None and np.issubdtype(dtype, attr_type):
            return dtype
        return self.default_types.get(self.datatypes.get(attr))

    def get_kind(self, attr: str, value: str='scalar') -> str:
        """Return if the attribute is of data type. Data type is 2- or
        3-dimensional and has the same data shape as `data`.

        Args:
            attr : The data attribute.
            value : value which is returned if the `attr` is not found.

        Returns:
            True if `attr` is of data type.
        """
        return self.kinds.get(attr, value)

    def get_ndim(self, attr: str, value: int=0) -> Tuple[int, ...]:
        """Return the acceptable number of dimenstions that the attribute's data
        may have.

        Args:
            attr : The data attribute.
            value : value which is returned if the `attr` is not found.

        Returns:
            Number of dimensions acceptable for the attribute.
        """
        return self.known_ndims.get((self.get_kind(attr)), value)

    def cast(self, attr: str, array: np.ndarray) -> np.ndarray:
        """Cast the attribute's data to the right data type.

        Args:
            attr : The data attribute.
            array : The attribute's data.

        Returns:
            Data array casted to the right data type.
        """
        return np.asarray(array, dtype=self.get_dtype(attr, array.dtype))

    @staticmethod
    def read_dataset_shapes(cxi_path: str, cxi_file: h5py.File) -> Dict[str, Tuple[int, ...]]:
        """Visit recursively all the underlying datasets and return
        their names and shapes.

        Args:
            cxi_path : Path of the HDF5 group.
            cxi_obj: Group object.

        Returns:
            List of all the datasets and their shapes inside `cxi_obj`.
        """
        shapes = {}

        def caller(sub_path, obj):
            if isinstance(obj, h5py.Dataset):
                shapes[os.path.join(cxi_path, sub_path)] = obj.shape

        if cxi_path in cxi_file:
            cxi_obj = cxi_file[cxi_path]
            if isinstance(cxi_obj, h5py.Dataset):
                shapes[cxi_path] = cxi_obj.shape
            elif isinstance(cxi_obj, h5py.Group):
                cxi_obj.visititems(caller)
            else:
                raise ValueError(f"Invalid CXI object at '{cxi_path:s}'")

        return shapes

    def read_attribute_shapes(self, attr: str, cxi_file: h5py.File) -> Dict[str, Tuple[int, ...]]:
        """Return a shape of the dataset containing the attribute's data inside
        a file.

        Args:
            attr : Attribute's name.
            cxi_file : HDF5 file object.

        Returns:
            List of all the datasets and their shapes inside `cxi_file`.
        """
        cxi_path = self.find_path(attr, cxi_file)
        return self.read_dataset_shapes(cxi_path, cxi_file)

    def read_attribute_indices(self, attr: str, cxi_files: List[h5py.File]) -> np.ndarray:
        """Return a set of indices of the dataset containing the attribute's data
        inside a set of files.

        Args:
            attr : Attribute's name.
            cxi_files : A list of HDF5 file objects.

        Returns:
            Dataset indices of the data pertined to the attribute `attr`.
        """
        files, cxi_paths, fidxs = [], [], []
        kind = self.get_kind(attr)

        for cxi_file in cxi_files:
            shapes = self.read_attribute_shapes(attr, cxi_file)
            for cxi_path, shape in shapes.items():
                if len(shape) not in self.get_ndim(attr):
                    err_txt = f'Dataset at {cxi_file.filename}:'\
                              f' {cxi_path} has invalid shape: {str(shape)}'
                    raise ValueError(err_txt)

                if kind in ['stack', 'sequence']:
                    files.extend(np.repeat(cxi_file.filename, shape[0]).tolist())
                    cxi_paths.extend(np.repeat(cxi_path, shape[0]).tolist())
                    fidxs.extend(np.arange(shape[0]).tolist())
                if kind in ['frame', 'scalar']:
                    files.append(cxi_file.filename)
                    cxi_paths.append(cxi_path)
                    fidxs.append(tuple())

        return np.array([files, cxi_paths, fidxs], dtype=object).T

def initializer(read_worker: Callable[[np.ndarray, Indices, Indices], np.ndarray],
                ss_idxs: np.ndarray, fs_idxs: np.ndarray):
    global worker
    worker = partial(read_worker, ss_idxs=ss_idxs, fs_idxs=fs_idxs)

def read_frame(index: np.ndarray) -> np.ndarray:
    return worker(index)

class CXIStore():
    """File handler class for HDF5 and CXI files. Provides an interface to
    save and load data attributes to a file. Support multiple files. The handler
    saves data to the first file.

    Attributes:
        file_dict : Dictionary of paths to the files and their file
            objects.
        file : :class:`h5py.File` object of the first file.
        protocol : :class:`pyrost.CXIProtocol` protocol object.
        mode : File mode. Valid modes are:

            * 'r' : Readonly, file must exist (default).
            * 'r+' : Read/write, file must exist.
            * 'w' : Create file, truncate if exists.
            * 'w-' or 'x' : Create file, fail if exists.
            * 'a' : Read/write if exists, create otherwise.
    """

    file_dict : Dict[str, h5py.File]

    def __init__(self, names: Union[str, List[str]], mode: str='r',
                 protocol: CXIProtocol=CXIProtocol.import_default()) -> None:
        """
        Args:
            names : Paths to the files.
            mode : Mode in which to open file; one of ('w', 'r', 'r+', 'a', 'w-').
            protocol : CXI protocol. Uses the default protocol if not provided.
        """
        if mode not in ['r', 'r+', 'w', 'w-', 'x', 'a']:
            raise ValueError(f'Wrong file mode: {mode}')
        names = protocol.str_to_list(names)

        self.file_dict = {data_file: None for data_file in names}
        self.protocol = protocol
        self.mode = mode

        with self:
            self.update_indices()

    def __bool__(self) -> bool:
        isopen = True
        for cxi_file in self.files():
            isopen &= bool(cxi_file)
        return isopen

    def __contains__(self, attr: str) -> bool:
        return attr in self._indices

    def __iter__(self) -> Iterator:
        return self._indices.__iter__()

    def __repr__(self) -> str:
        return self._indices.__repr__()

    def __str__(self) -> str:
        return self._indices.__str__()

    def __enter__(self) -> CXIStore:
        self.open()
        return self

    def __exit__(self, exc_type: Optional[BaseException], exc: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        self.close()

    def update_indices(self) -> None:
        """Read the files for the data attributes contained in the protocol."""
        indices = {}
        if self:
            for attr in self.protocol:
                kind = self.protocol.get_kind(attr)
                try:
                    if kind in ['stack', 'sequence', 'scalar']:
                        idxs = self.protocol.read_attribute_indices(attr, self.files())
                    if kind == 'frame':
                        idxs = self.protocol.read_attribute_indices(attr, self.files())
                except ValueError as err:
                    print(f'{attr:s} is not loaded: {err}')
                else:
                    if idxs.size:
                        indices[attr] = idxs

        self._indices = indices

    @property
    def file(self) -> h5py.File:
        return self.files()[0]

    def open(self) -> None:
        """Open the files."""
        if not self:
            for data_file in self.file_dict:
                self.file_dict[data_file] = h5py.File(data_file, mode=self.mode)

    def close(self) -> None:
        """Close the files."""
        for cxi_file in self.files():
            cxi_file.close()

    def filenames(self) -> List[str]:
        """Return a list of paths to the files.

        Returns:
            List of paths to the files.
        """
        return list(self.file_dict.keys())

    def files(self) -> List[h5py.File]:
        """Return a list of file objects of the files.

        Returns:
            List of file objects of the files.
        """
        return list(self.file_dict.values())

    def indices(self) -> np.ndarray:
        """Return a list of frame indices of the data contained in the
        files.

        Returns:
            A list of frame indices of the data in the files.
        """
        return np.arange(self._indices.get('data', np.array([])).shape[0])

    def keys(self) -> KeysView:
        """Return a set of the attribute's names contained in the files.

        Returns:
            A set of the attribute's names in the files.
        """
        return self._indices.keys()

    def values(self) -> ValuesView:
        """Return a set of the attribute's indices in the files.

        Returns:
            A set of the attribute's indices in the files.
        """
        return self._indices.values()

    def items(self) -> ItemsView:
        """Return a set of `(name, index)` pairs. `name` is the attribute's name
        contained in the files, and `index` it's corresponding set of indices.

        Returns:
            A set of `(name, index)` pairs of the data attributes in the files.
        """
        return self._indices.items()

    def read_shape(self) -> Tuple[int, int]:
        """Read the input files and return a shape of the `frame` type data attribute.

        Returns:
            The shape of the 2D `frame`-like data attribute.
        """
        if self:
            for attr, indices in self._indices.items():
                kind = self.protocol.get_kind(attr)
                if kind in ['stack', 'frame']:
                    return self.file_dict[indices[0, 0]][indices[0, 1]].shape[-2:]

            return (0, 0)

        raise ValueError('Invalid file objects: the file is closed')

    @staticmethod
    def _read_worker_sequence(index: np.ndarray) -> np.ndarray:
        return h5py.File(index[0])[index[1]][index[2]]

    @staticmethod
    def _read_worker_frame(index: np.ndarray, ss_idxs: Indices, fs_idxs: Indices) -> np.ndarray:
        return h5py.File(index[0])[index[1]][index[2]][..., ss_idxs, fs_idxs]

    def _load_stack(self, attr: str, idxs: Optional[Indices], ss_idxs: Indices,
                    fs_idxs: Indices, processes: int, verbose: bool) -> np.ndarray:
        stack = []
        if idxs is None:
            idxs = self.indices()
        idxs = np.atleast_1d(idxs)

        with Pool(processes=processes, initializer=initializer,
                  initargs=(type(self)._read_worker_frame, ss_idxs, fs_idxs)) as pool:
            for frame in tqdm(pool.imap(read_frame, self._indices[attr][idxs]),
                              total=self.indices()[idxs].size, disable=not verbose,
                              desc=f'Loading {attr:s}'):
                stack.append(frame)

        return self.protocol.cast(attr, np.stack(stack, axis=0))

    def _load_frame(self, attr: str, ss_idxs: Indices, fs_idxs: Indices) -> np.ndarray:
        dset = self._read_worker_frame(self._indices[attr][0], ss_idxs, fs_idxs)
        return self.protocol.cast(attr, dset)

    def _load_sequence(self, attr: str, idxs: Optional[Indices]) -> np.ndarray:
        sequence = []
        if idxs is None:
            idxs = self.indices()
        idxs = np.atleast_1d(idxs)

        for index in self._indices[attr][idxs]:
            sequence.append(self._read_worker_sequence(index))

        return self.protocol.cast(attr, np.array(sequence))

    def load_attribute(self, attr: str, idxs: Optional[Indices]=None, processes: int=1,
                       ss_idxs: Indices=slice(None), fs_idxs: Indices=slice(None), verbose: bool=True) -> np.ndarray:
        """Load a data attribute from the files.

        Args:
            attr : Attribute's name to load.
            idxs : A list of frames' indices to load.
            processes : Number of parallel workers used during the loading.
            ss_idxs : Slow (vertical) axis indices used to load the data attributes of
                `frame` and `stack` types.
            fs_idxs : Fast (horizontal) axis indices used to lead the data attributes
                of `frame` and `stack` types.
            verbose : Set the verbosity of the loading process.

        Raises:
            ValueError : If the attribute's kind is invalid.

        Returns:
            Attribute's data array.
        """
        kind = self.protocol.get_kind(attr)

        if kind not in ['stack', 'frame', 'scalar', 'sequence']:
            raise ValueError(f'Invalid kind: {kind:s}')

        if self:
            if kind == 'stack':
                return self._load_stack(attr=attr, idxs=idxs, processes=processes,
                                        ss_idxs=ss_idxs, fs_idxs=fs_idxs, verbose=verbose)
            if kind == 'frame':
                return self._load_frame(attr=attr, ss_idxs=ss_idxs, fs_idxs=fs_idxs, )
            if kind == 'scalar':
                return self._load_sequence(attr, idxs=0)
            if kind == 'sequence':
                return self._load_sequence(attr=attr, idxs=idxs)

        raise ValueError('Invalid file objects: the file is closed')

    def find_dataset(self, attr: str) -> str:
        """Return the path to the attribute from the first file. Return the default
        path if the attribute is not found inside the first file.

        Args:
            attr : Attribute's name.

        Returns:
            Path to the attribute inside the first file.
        """
        cxi_path = self.protocol.find_path(attr, self.file)

        if cxi_path:
            return cxi_path
        return self.protocol.get_load_paths(attr)[0]

    def _save_stack(self, attr: str, data: np.ndarray, mode: str='overwrite',
                    idxs: Optional[Indices]=None) -> None:
        cxi_path = self.find_dataset(attr)

        if cxi_path in self.file and self.file[cxi_path].shape[1:] == data.shape[1:]:
            if mode == 'append':
                self.file[cxi_path].resize(self.file[cxi_path].shape[0] + data.shape[0],
                                               axis=0)
                self.file[cxi_path][-data.shape[0]:] = data
            elif mode == 'overwrite':
                self.file[cxi_path].resize(data.shape[0], axis=0)
                self.file[cxi_path][...] = data
            elif mode == 'insert':
                if idxs is None or len(idxs) != data.shape[0]:
                    raise ValueError('Incompatible indices')
                self.file[cxi_path].resize(max(self.file[cxi_path].shape[0], max(idxs) + 1),
                                               axis=0)
                self.file[cxi_path][idxs] = data

        else:
            if cxi_path in self.file:
                del self.file[cxi_path]
            self.file.create_dataset(cxi_path, data=data, shape=data.shape,
                                         chunks=(1,) + data.shape[1:],
                                         maxshape=(None,) + data.shape[1:],
                                         dtype=self.protocol.get_dtype(attr, data.dtype))

    def _save_data(self, attr: str, data: np.ndarray) -> None:
        cxi_path = self.find_dataset(attr)

        if cxi_path in self.file and self.file[cxi_path].shape == data.shape:
            self.file[cxi_path][...] = data

        else:
            if cxi_path in self.file:
                del self.file[cxi_path]
            self.file.create_dataset(cxi_path, data=data, shape=data.shape,
                                         dtype=self.protocol.get_dtype(attr, data.dtype))

    def save_attribute(self, attr: str, data: np.ndarray, mode: str='overwrite',
                       idxs: Optional[Iterable[int]]=None) -> None:
        """Save a data array pertained to the data attribute into the first file.

        Args:
            attr : Attribute's name.
            data : Data array.
            mode : Writing mode:

                * `append` : Append the data array to already existing dataset.
                * `insert` : Insert the data under the given indices `idxs`.
                * `overwrite` : Overwrite the existing dataset.

            idxs : Indices where the data is saved. Used only if `mode` is set to
                'insert'.

        Raises:
            ValueError : If the attribute's kind is invalid.
        """
        if self.mode == 'r':
            raise ValueError('File is open in read-only mode')
        kind = self.protocol.get_kind(attr)

        if self:
            if kind in ['stack', 'sequence']:
                return self._save_stack(attr=attr, data=data, mode=mode, idxs=idxs)

            if kind in ['frame', 'scalar']:
                return self._save_data(attr=attr, data=data)

            raise ValueError(f'Invalid kind: {kind:s}')

        raise ValueError('Invalid file objects: the file is closed')



""":class:`pyrost.SpeckleTracking` provides an interface to perform the reference
image and lens wavefront reconstruction and offers two methods
(:func:`pyrost.SpeckleTracking.train`, :func:`pyrost.SpeckleTracking.train_adapt`)
to perform the iterative R-PXST update until the error metric converges to a minimum.
"""
from __future__ import annotations
from typing import Dict, List, Optional, Tuple, Union
from weakref import ReferenceType
from tqdm.auto import tqdm
import numpy as np
from .data_container import DataContainer, dict_to_object
from .bfgs import BFGS
from .bin import (gaussian_filter, KR_reference, LOWESS_reference, pm_gsearch,
                  pm_rsearch, pm_devolution, tr_gsearch, pm_errors, pm_total_error,
                  ref_errors, ct_integrate)

class SpeckleTracking(DataContainer):
    """Wrapper class for the  Robust Speckle Tracking algorithm.
    Provides an interface to perform the reference image and lens
    wavefront reconstruction.

    Attributes:
        attr_set : Set of attributes in the container which are necessary
            to initialize in the constructor.
        init_set : Set of optional data attributes.

    Notes:
        **Necessary attributes**:

        * data : Measured frames.
        * di_pix : The sample's translations along the vertical axis in
          pixels.
        * dj_pix : The sample's translations along the horizontal axis in
          pixels.
        * ds_y : `reference_image` sampling interval in pixels along
          the vertical axis.
        * ds_x : `reference_image` sampling interval in pixels along
          the horizontal axis.
        * num_threads : Specify number of threads that are used in all the
          calculations.
        * parent : The parent :class:`STData` container.
        * pixel_map : The pixel mapping between the data at the detector's
          plane and the reference image at the reference plane.
        * scale_map : Huber scaling map.
        * whitefield : Measured frames' white-field.

        **Optional attributes**:

        * error : Average error of the reference image and pixel mapping fit.
        * hval : Smoothing kernel bandwidth used in `reference_image`
          regression. The value is given in pixels.
        * m0 : The lower bounds of the horizontal detector axis of
          the reference image at the reference frame in pixels.
        * n0 : The lower bounds of the vertical detector axis of
          the reference image at the reference frame in pixels.
        * reference_image : The unabberated reference image of the sample.

    See Also:

        * :func:`pyrost.bin.KR_reference` : Full details of the `reference_image`
          update using kernel regression.
        * :func:`pyrost.bin.LOWESS_reference` : Full details of the `reference_image`
          update using local weighted linear regression.
        * :func:`pyrost.bin.pm_gsearch` : Full details of the `pixel_map` grid search
          update.

    """
    attr_set = {'data', 'dj_pix', 'di_pix', 'ds_y', 'ds_x', 'num_threads', 'pixel_map',
                'parent', 'whitefield'}
    init_set = {'error', 'hval', 'initial', 'ref_orig', 'reference_image',
                'scale_map', 'test_mask', 'train_mask'}

    dtypes_32 = {'data': np.uint32, 'dij_pix': np.float32, 'pixel_map': np.float32,
                 'scale_map': np.float32, 'whitefield': np.float32}
    dtypes_64 = {'data': np.uint64, 'dij_pix': np.float64, 'pixel_map': np.float64,
                 'scale_map': np.float64, 'whitefield': np.float64}

    # Necessary attributes
    data : np.ndarray
    dj_pix : np.ndarray
    di_pix : np.ndarray
    ds_x : float
    ds_y : float
    num_threads : int
    parent : ReferenceType
    pixel_map : np.ndarray
    whitefield : np.ndarray

    # Automatically generated attributes
    reference_image : np.ndarray
    ref_orig : np.ndarray
    scale_map : np.ndarray
    test_mask : np.ndarray
    train_mask : np.ndarray

    # Optional attributes
    hval : Optional[float]
    error : Optional[float]
    initial : Optional[SpeckleTracking]

    def __init__(self, parent: ReferenceType, **kwargs: Union[int, float, np.ndarray]) -> None:
        """
        Args:
            parent : The Speckle tracking data container, from which the
                object is derived.
            kwargs : Dictionary of the attributes' data specified in `attr_set`
                and `init_set`.

        Raises:
            ValueError : If an attribute specified in `attr_set` has not been
                provided.
        """
        super(SpeckleTracking, self).__init__(parent=parent, **kwargs)
        self._init_functions(test_mask=self._test_mask, train_mask=lambda: ~self.test_mask,
                             ref_orig=self._ref_orig, reference_image=self._reference_image,
                             scale_map=lambda: np.sqrt(self.whitefield))
        self._init_attributes()

    def _ref_orig(self) -> np.ndarray:
        y_orig = self.di_pix.max() - self.pixel_map[0].min()
        x_orig = self.dj_pix.max() - self.pixel_map[1].min()
        return np.array([y_orig, x_orig]).astype(int)

    def _reference_image(self) -> np.ndarray:
        shape = self.pixel_map.max(axis=(1, 2)) - np.array([self.di_pix.min(), self.dj_pix.min()])
        shape = ((shape + self.ref_orig) / np.array([self.ds_y, self.ds_x]))
        return np.ones(shape.astype(int) + 1, dtype=self.whitefield.dtype)

    def _test_mask(self, test_ratio: float=0.2) -> np.ndarray:
        idxs = np.random.choice(self.whitefield.size, size=int(self.whitefield.size * test_ratio),
                                replace=False)
        idxs = np.unravel_index(idxs, self.whitefield.shape)
        test_mask = np.zeros(self.whitefield.shape, dtype=bool)
        test_mask[idxs] = True
        return test_mask

    def __repr__(self) -> str:
        with np.printoptions(threshold=6, edgeitems=2, suppress=True, precision=3):
            return {key: val.ravel() if isinstance(val, np.ndarray) else val
                    for key, val in self.items()}.__repr__()

    def __str__(self) -> str:
        with np.printoptions(threshold=6, edgeitems=2, suppress=True, precision=3):
            return {key: val.ravel() if isinstance(val, np.ndarray) else val
                    for key, val in self.items()}.__str__()

    @dict_to_object
    def create_initial(self) -> SpeckleTracking:
        """Create a :class:`SpeckleTracking` object with preliminary approximation of
        the pixel mapping and reference profile. The object is saved into `initial`
        attribute. Necessary to calculate normalized error metrics.

        Returns:
            A new :class:`SpeckleTracking` object with the preliminary
            :class:`SpeckleTracking` object saved in the `initial` attribute.
        """
        initial = self.parent().get_st(ds_x=self.ds_x, ds_y=self.ds_y).update_errors()
        return {'initial': initial}

    @dict_to_object
    def test_train_split(self, test_ratio: float=0.1) -> SpeckleTracking:
        """Update test / train subsets split. Required to calculate the Cross-validation
        error metric.

        Args:
            test_ratio : Ratio between the size of the test subset and the whole dataset.

        Returns:
            A new :class:`SpeckleTracking` object with a new test / train subsets split.

        See Also:
            :func:`SpeckleTracking.CV` : Full details on the Cross-validation error
            metric.
        """
        test_mask = self._test_mask(test_ratio)
        return {'test_mask': test_mask, 'train_mask': ~test_mask}

    @dict_to_object
    def update_reference(self, hval: float, method: str='KerReg') -> SpeckleTracking:
        r"""Return a new :class:`SpeckleTracking` object with the updated
        `reference_image`. The reference profile is estimated either by
        Kernel regression ('KerReg') [KerReg]_ or Local weighted linear\
        egressin ('LOWESS') [LOWESS]_.

        Args:
            hval : Smoothing kernel bandwidth used in `reference_image`
                regression. The value is given in pixels.
            method : `reference_image` update algorithm. The following keyword
                values are allowed:

                * `KerReg` : Kernel regression algorithm.
                * `LOWESS` : Local weighted linear regression.

        Raises:
            ValueError : If `method` keyword value is not valid.

        Returns:
            A new :class:`SpeckleTracking` object with the updated
            `reference_image`.

        Notes:
            The reference profile estimator is given by:

            .. math::
                I_\text{ref}(f_x i, f_y j) = \frac{\sum_n \sum_{i^{\prime}}
                \sum_{j^{\prime}} K(f_x i - u^x_{i^{\prime}j^{\prime}} + \Delta i_n,
                f_y j - u^y_{i^{\prime}j^{\prime}} + \Delta j_n, h) \;
                W_{i^{\prime}j^{\prime}} I_{n i^{\prime} j^{\prime}}}
                {\sum_n \sum_{i^{\prime}} \sum_{j^{\prime}}
                K(f_x i - u^x_{i^{\prime}j^{\prime}} + \Delta i_n,
                f_y j - u^y_{i^{\prime}j^{\prime}} + \Delta j_n, h) \;
                W^2_{i^{\prime} j^{\prime}}},

            where :math:`K(i, j, h) = \exp(\frac{i\:\delta u + j\:\delta v}{h})
            / \sqrt{2 \pi}` is the Gaussian kernel, :math:`u^x_{ij}, u^y_{ij}`
            are the horizontal and vertical components of the pixel mapping,
            :math:`\Delta i_n, \Delta j_n` are the sample translation along the
            horizontal and vertical axes in pixels, :math:`I_{nij}` are the measured
            stack of frames, and `W_{ij}` is the white-field.

        See Also:
            :func:`pyrost.bin.make_reference` : Full details of the `reference_image`
                update algorithm.

        References:
            .. [KerReg] E. A. Nadaraya, “On estimating regression,” Theory Probab. & Its
                       Appl. 9, 141-142 (1964).

            .. [LOWESS] H.-G. Müller, “Weighted local regression and kernel methods for
                       nonparametric curve fitting,” J. Am. Stat. Assoc. 82, 231-238
                       (1987).
        """
        if method == 'KerReg':
            I0, n0, m0 = KR_reference(I_n=self.data, W=self.whitefield, u=self.pixel_map,
                                      di=self.di_pix, dj=self.dj_pix, ds_y=self.ds_y,
                                      ds_x=self.ds_x, hval=hval, num_threads=self.num_threads)
        elif method == 'LOWESS':
            I0, n0, m0 = LOWESS_reference(I_n=self.data, W=self.whitefield, u=self.pixel_map,
                                          di=self.di_pix, dj=self.dj_pix, ds_y=self.ds_y,
                                          ds_x=self.ds_x, hval=hval, num_threads=self.num_threads)
        else:
            raise ValueError('Method keyword is invalid')
        return {'hval': hval, 'ref_orig': np.array([n0, m0]), 'reference_image': I0}

    def ref_indices(self) -> np.ndarray:
        """Return an array of reference profile pixel indices.

        Returns:
            An array of reference profile pixel indices.
        """
        idxs = np.indices(self.reference_image.shape) 
        idxs[0] = idxs[0] * self.ds_y - self.ref_orig[0]
        idxs[1] = idxs[1] * self.ds_x - self.ref_orig[1]
        return idxs

    @dict_to_object
    def update_pixel_map(self, search_window: Tuple[float, float, float], blur: float=0.0,
                         integrate: bool=False, method: str='gsearch',
                         extra_args: Dict[str, Union[int, float]]={}) -> SpeckleTracking:
        r"""Return a new :class:`SpeckleTracking` object with the updated pixel mapping
        (`pixel_map`) and Huber scale mapping (`scale_map`). The update is performed
        with the adaptive Huber regression [HUBER]_. The Huber loss function is minimized
        by the non-gradient methods enlisted in the `method` argument.

        Args:
            search_window : A tuple of three elements ('sw_y', 'sw_x', 'sw_s'). The elements
                are the following:

                * `sw_y` : Search window size in pixels along the horizontal detector
                  axis.
                * `sw_x` : Search window size in pixels along the vertical detector
                  axis.
                * `sw_s` : Search window size of the Huber scaling map. Given as a ratio
                  (0.0 - 1.0) relative to the scaling map value before the update.

            blur : Smoothing kernel bandwidth used in `pixel_map` regularisation.
                The value is given in pixels.
            integrate : Ensure that the updated pixel map is irrotational by
                integrating and taking the derivative.
            method : `pixel_map` update algorithm. The following keyword
                values are allowed:

                * `gsearch` : Grid search algorithm.
                * `rsearch` : Random search algorithm.
                * `de`      : Differential evolution algorithm.

            extra_args : Extra arguments for pixel map update methods. Accepts the
                following keyword arguments:

                * `grid_size` : Grid size along one of the detector axes for
                  the 'gsearch' method. The grid shape is then (grid_size,
                  grid_size). The default value is :code:`int(0.5 * (sw_x + sw_y))`.
                * `n_trials` : Number of points generated at each  pixel in
                  the detector grid for the 'rsearch' method. The default value
                  is :code:`int(0.25 * (sw_x + sw_y)**2)`.
                * `n_iter` : The maximum number of generations over which the
                  entire population is evolved in the 'de' method. The default
                  value is 5.
                * `pop_size` : The total population size in the 'de' method.
                  Must be greater or equal to 4. The default value is
                  :code:`max(int(0.25 * (sw_x + sw_y)**2) / n_iter, 4)`.
                * `mutation` : The mutation constant in the 'de' method. It should
                  be in the range [0, 2]. The default value is 0.75.
                * `recombination` : The recombination constant  in the 'de' method,
                  should be in the range [0, 1]. The default value is 0.7.
                * `seed` : Specify seed for the random number generation.
                  Generated automatically if not provided.

        Raises:
            AttributeError : If `reference_image` was not generated beforehand.
            ValueError : If `method` keyword value is not valid.

        Returns:
            A new :class:`SpeckleTracking` object with the updated `pixel_map` and
            `scale_map`.

        Notes:
            The pixel mapping update is carried out separately at each pixel
            :math:`{i, j}` in the detector grid by minimizing the Huber error metric
            given by:

            .. math::

                \varepsilon_{ij}(\delta i, \delta j, s) = \frac{1}{N}
                \sum_{n = 1}^N \left[ s + \mathcal{H}_{1.35} \left( \frac{I_{nij} -
                W_{ij} I_\text{ref}(f_x i - u^x_{ij} - \delta i + \Delta i_n,
                f_y i - u^y_{ij} - \delta j + \Delta j_n)}{s} \right) s \right],

            where :math:`I_\text{ref}` is the reference profile, :math:`u^x_{ij},
            u^y_{ij}` are the horizontal and vertical components of the pixel mapping,
            :math:`\Delta i_n, \Delta j_n` are the sample translation along the
            horizontal and vertical axes in pixels, :math:`I_{nij}` are the measured
            stack of frames, and `W_{ij}` is the white-field.

        See Also:

            * :func:`pyrost.bin.pm_gsearch` : Full details of the grid search
              update method.
            * :func:`pyrost.bin.pm_rsearch` : Full details of the random search
              update method.
            * :func:`pyrost.bin.pm_devolution` : Full details of the differential
              evolution update method.

        References:
            .. [HUBER] A. B. Owen, “A robust hybrid of lasso and ridge regression,”
                      (2006).
        """
        n_iter = extra_args.get('n_iter', 5)
        mutation = extra_args.get('mutation', 0.75)
        recombination = extra_args.get('recombination', 0.7)
        seed = extra_args.get('seed', np.random.default_rng().integers(0, np.iinfo(np.int_).max,
                                                                       endpoint=False))
        grid_size = extra_args.get('grid_size', (int(0.5 * search_window[0]) + 1,
                                                 int(0.5 * search_window[1]) + 1,
                                                 int(50.0 * search_window[2]) + 1))
        n_trials = extra_args.get('n_trials', max(np.prod(grid_size), 2))
        pop_size = extra_args.get('pop_size', max(n_trials / n_iter, 4))

        if method == 'gsearch':
            pm, scale, derr = pm_gsearch(I_n=self.data, W=self.whitefield, I0=self.reference_image,
                                         u0=self.pixel_map, di=self.di_pix - self.ref_orig[0],
                                         dj=self.dj_pix - self.ref_orig[1], search_window=search_window,
                                         grid_size=grid_size, ds_y=self.ds_y, ds_x=self.ds_x,
                                         sigma=self.scale_map, num_threads=self.num_threads)
        elif method == 'rsearch':
            pm, scale, derr = pm_rsearch(I_n=self.data, W=self.whitefield, I0=self.reference_image,
                                         u0=self.pixel_map, di=self.di_pix - self.ref_orig[0],
                                         dj=self.dj_pix - self.ref_orig[1], search_window=search_window,
                                         n_trials=n_trials, seed=seed, ds_y=self.ds_y, ds_x=self.ds_x,
                                         sigma=self.scale_map, num_threads=self.num_threads)
        elif method == 'de':
            pm, scale, derr = pm_devolution(I_n=self.data, W=self.whitefield, I0=self.reference_image,
                                            u0=self.pixel_map, di=self.di_pix - self.ref_orig[0],
                                            dj=self.dj_pix - self.ref_orig[1], search_window=search_window,
                                            pop_size=pop_size, n_iter=n_iter, seed=seed,
                                            ds_y=self.ds_y, ds_x=self.ds_x, sigma=self.scale_map,
                                            F=mutation, CR=recombination, num_threads=self.num_threads)
        else:
            raise ValueError('Method keyword is invalid')

        dpm = pm - self.pixel_map
        std = derr.std()
        if std:
            derr = (derr - derr.min() + 1.0) / std
        else:
            derr = (derr - derr.min() + 1.0)
        if blur > 0.0:
            norm = gaussian_filter(derr, (blur, blur))
            dpm = gaussian_filter(dpm * derr, (0, blur, blur),
                                  num_threads=self.num_threads) / norm

        uy_avg, ux_avg = dpm.mean(axis=(1, 2))
        pm[0] = (dpm[0] - uy_avg) + self.pixel_map[0]
        pm[1] = (dpm[1] - ux_avg) + self.pixel_map[1]

        if integrate:
            phi = ct_integrate(sy_arr=pm[0], sx_arr=pm[1], num_threads=self.num_threads)
            pm[0] = np.gradient(phi, axis=0)
            pm[1] = np.gradient(phi, axis=1)

        return {'pixel_map': pm, 'scale_map': scale}

    @dict_to_object
    def update_errors(self) -> SpeckleTracking:
        """Return a new :class:`SpeckleTracking` object with
        the updated mean Huber error metric `error`.

        Returns:
            A new :class:`SpeckleTracking` object with the updated
            `error`.

        Raises:
            AttributeError : If `reference_image` was not generated
                beforehand.

        See Also:
            * :func:`pyrost.bin.ref_errors` : Full details of the reference
              update error metric.
            * :func:`pyrost.bin.pm_errors` : Full details of the pixel
              mapping update error metric.
        """
        error = pm_total_error(I_n=self.data, W=self.whitefield,
                               I0=self.reference_image, u=self.pixel_map,
                               di=self.di_pix - self.ref_orig[0], ds_y=self.ds_y,
                               dj=self.dj_pix - self.ref_orig[1], ds_x=self.ds_x,
                               sigma=self.scale_map, num_threads=self.num_threads)
        if self.initial is None:
            return {'error': error}
        return {'error': error / self.initial.error}

    @dict_to_object
    def update_translations(self, sw_x: float, sw_y: float=0.0, blur=0.0) -> SpeckleTracking:
        """Return a new :class:`SpeckleTracking` object with the updated sample
        pixel translations (`di_pix`, `dj_pix`). The update is performed with the
        adaptive Huber regression [HUBER]_. The Huber loss function is minimized
        by the grid search algorithm.

        Args:
            sw_x : Search window size in pixels along the horizontal detector
                axis.
            sw_y : Search window size in pixels along the vertical detector
                axis.
            blur : Smoothing kernel bandwidth used in `dj_pix` and `di_pix`
                post-update. The value is given in pixels.

        Returns:
            A new :class:`SpeckleTracking` object with the updated
            `di_pix`, `dj_pix`.

        Raises:
            AttributeError : If `reference_image` was not generated beforehand.

        See Also:
            :func:`pyrost.bin.tr_gsearch` : Full details of the sample translations
                update algorithm.
        """
        dij = tr_gsearch(I_n=self.data, W=self.whitefield, I0=self.reference_image,
                         u=self.pixel_map, di=self.di_pix - self.ref_orig[0],
                         dj=self.dj_pix - self.ref_orig[1], sw_y=sw_y, sw_x=sw_x,
                         ds_y=self.ds_y, ds_x=self.ds_x, sigma=self.scale_map,
                         num_threads=self.num_threads)
        dij += self.ref_orig

        di_pix = np.ascontiguousarray(dij[:, 0])
        dj_pix = np.ascontiguousarray(dij[:, 1])
        if blur > 0.0:
            di_pix = gaussian_filter(di_pix - self.di_pix, blur, mode='nearest',
                                      num_threads=self.num_threads) + self.di_pix
            dj_pix = gaussian_filter(dj_pix - self.dj_pix, blur, mode='nearest',
                                      num_threads=self.num_threads) + self.dj_pix

        return {'di_pix': di_pix, 'dj_pix': dj_pix}

    def error_profile(self, kind: str='pixel_map') -> np.ndarray:
        """Return a error metrics for the reference profile and pixel mapping
        updates. The error metrics may be normalized by the error of the initial
        estimations of the reference profile and pixel mapping function. The
        normalization is performed if :func:`SpeckleTracking.create_initial` was
        invoked before.

        Args:
            kind : Choose between generating the error metric of the pixel mapping
                ('pixel_map') or reference image ('reference')update.

        Raises:
            ValueError : If `kind` keyword value is not valid.

        Returns:
            Residual profile.

        See Also:
            :func:`SpeckleTracking.update_reference` : More details on the reference
            profile update procedure.
            :func:`SpeckleTracking.update_pixel_map` : More details on the pixel
            mapping update procedure.
        """

        if kind == 'pixel_map':
            pm_err = pm_errors(I_n=self.data, W=self.whitefield, u=self.pixel_map,
                               I0=self.reference_image, sigma=self.scale_map,
                               di=self.di_pix - self.ref_orig[0], ds_y=self.ds_y,
                               dj=self.dj_pix - self.ref_orig[1], ds_x=self.ds_x,
                               num_threads=self.num_threads)
            if self.initial is None:
                return pm_err
            pm_tot = self.initial.error_profile(kind=kind)
            return np.where(pm_tot, pm_err / pm_tot, 0.0)

        if kind == 'reference':
            ref_err = ref_errors(I_n=self.data, W=self.whitefield, u=self.pixel_map,
                                 I0=self.reference_image, hval=self.hval,
                                 di=self.di_pix - self.ref_orig[0], ds_y=self.ds_y,
                                 dj=self.dj_pix - self.ref_orig[1], ds_x=self.ds_x,
                                 num_threads=self.num_threads)
            if self.initial is None:
                return ref_err
            self.initial.hval = self.hval
            ref_tot = self.initial.error_profile(kind=kind)

            idxs0 = np.moveaxis(self.initial.ref_indices(), 0, -1)
            idxs1 = np.moveaxis(self.ref_indices(), 0, -1)

            start = np.max((idxs0[0, 0], idxs1[0, 0]), axis=0)
            end = np.min((idxs0[-1, -1], idxs1[-1, -1]), axis=0)
            ps = 0.5 * np.array([self.ds_y, self.ds_x])

            roi0 = np.concatenate([np.argwhere(np.all(np.abs(idxs0 - start) < ps, axis=-1)),
                                   np.argwhere(np.all(np.abs(idxs0 - end) < ps, axis=-1)) + 1])
            roi1 = np.concatenate([np.argwhere(np.all(np.abs(idxs1 - start) < ps, axis=-1)),
                                   np.argwhere(np.all(np.abs(idxs1 - end) < ps, axis=-1)) + 1])

            ref_tot = ref_tot[roi0[0, 0]:roi0[1, 0], roi0[0, 1]:roi0[1, 1]]
            ref_err = ref_err[roi1[0, 0]:roi1[1, 0], roi1[0, 1]:roi1[1, 1]]
            return np.where(ref_tot, ref_err / ref_tot, 0.0)

        raise ValueError('kind keyword is invalid')

    def find_hopt(self, h0: float=1.0, method: str='KerReg',
                  epsilon: float=1e-3, maxiter: int=10, gtol: float=1e-5,
                  verbose: bool=False) -> float:
        """Find the optimal kernel bandwidth by finding the bandwidth, that minimized
        the Cross-validation error metric. The minimization process is performed with
        the quasi-Newton method of Broyden, Fletcher, Goldfarb, and Shanno [BFGS]_.

        Args:
            h0 : Initial guess of kernel bandwidth in pixels.
            method : `reference_image` update algorithm. The following keyword values
                are allowed:

                * `KerReg` : Kernel regression algorithm.
                * `LOWESS` : Local weighted linear regression.

            epsilon : The step size used in the estimation of the fitness gradient.
            maxiter : Maximum number of iterations in the minimization loop.
            gtol : Gradient norm must be less than `gtol` before successful
                termination.
            verbose : Print convergence message if True.

        Returns:
            Optimal kernel bandwidth in pixels.

        References:
            .. [BFGS] S. Wright, J. Nocedal et al., “Numerical optimization,” Springer
                     Sci. 35, 7 (1999).
        """
        optimizer = BFGS(lambda hval: self.CV(hval, method), np.atleast_1d(h0),
                         epsilon=epsilon)
        itor = tqdm(range(maxiter), disable=not verbose,
                    bar_format='{desc} {percentage:3.0f}% {bar} Iteration {n_fmt}'\
                               ' / {total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]')
        for _ in itor:
            optimizer.step()
            if verbose:
                itor.set_description(f"hopt = {optimizer.state_dict()['xk'].item():.3f}, " \
                                     f"gnorm = {optimizer.state_dict()['gnorm'].item():.3e}")
            if optimizer.state_dict()['gnorm'] < gtol:
                break
        return optimizer.state_dict()['xk'].item()

    def train_adapt(self, search_window: Tuple[float, float, float], h0: float,
                    blur: float=0.0, n_iter: int=30, f_tol: float=1e-8,
                    ref_method: str='KerReg', pm_method: str='rsearch',
                    pm_args: Dict[str, Union[bool, int, float, str]]={},
                    options: Dict[str, Union[bool, float, str]]={},
                    verbose: bool=True) -> Tuple[SpeckleTracking, Dict[str, List[float]]]:
        """Perform adaptive iterative Robust Speckle Tracking update. The reconstruction
        cycle consists of: (i) estimating an optimal kernel bandwidth for the reference
        image estimate (see :func:`SpeckleTracking.find_hopt`); (ii) generating the
        reference image (see :func:`SpeckleTracking.update_reference`); (iii) updating
        the pixel mapping between a stack of frames and the generated reference image
        (see :func:`SpeckleTracking.update_pixel_map`); (iv) updating the sample
        translation vectors (if needed, see :func:`SpeckleTracking.update_translations`);
        and (v) calculating the mean Huber error (see :func:`SpeckleTracking.update_error`).

        Args:
            search_window : A tuple of three elements ('sw_y', 'sw_x', 'sw_s'). The elements
                are the following:

                * `sw_y` : Search window size in pixels along the horizontal detector
                  axis.
                * `sw_x` : Search window size in pixels along the vertical detector
                  axis.
                * `sw_s` : Search window size of the Huber scaling map. Given as a ratio
                  (0.0 - 1.0) relative to the scaling map value before the update.

            h0 : Smoothing kernel bandwidth used in `reference_image` estimation.
                The value is given in pixels.
            blur : Smoothing kernel bandwidth used in `pixel_map` regularisation.
                The value is given in pixels.
            n_iter : Maximum number of iterations.
            f_tol : Tolerance for termination by the change of the average error. The
                iteration stops when ``(error^k - error^{k+1})/max{|error^k|, |error^{k+1}|,
                1} <= f_tol``.
            ref_method : `reference_image` update algorithm. The following
                keyword values are allowed:

                * `KerReg` : Kernel regression algorithm.
                * `LOWESS` : Local weighted linear regression.

            pm_method : `pixel_map` update algorithm. The following keyword
                values are allowed:

                * `gsearch` : Grid search algorithm.
                * `rsearch` : Random search algorithm.
                * `de`      : Differential evolution algorithm.

            pm_args : Pixel map update options. Accepts the following keyword
                arguments:

                * `integrate` : Ensure that the updated pixel map is irrotational
                  by integrating and taking the derivative. False by default.
                * `grid_size` : Grid size along one of the detector axes for
                  the 'gsearch' method. The grid shape is then (grid_size,
                  grid_size). The default value is :code:`int(0.5 * (sw_x + sw_y))`.
                * `n_trials` : Number of points generated at each  pixel in
                  the detector grid for the 'rsearch' method. The default value
                  is :code:`int(0.25 * (sw_x + sw_y)**2)`.
                * `n_iter` : The maximum number of generations over which the
                  entire population is evolved in the 'de' method. The default
                  value is 5.
                * `pop_size` : The total population size in the 'de' method.
                  Must be greater or equal to 4. The default value is
                  :code:`max(int(0.25 * (sw_x + sw_y)**2) / n_iter, 4)`.
                * `mutation` : The mutation constant in the 'de' method. It should
                  be in the range [0, 2]. The default value is 0.75.
                * `recombination` : The recombination constant  in the 'de' method,
                  should be in the range [0, 1]. The default value is 0.7.
                * `seed` : Specify seed for the random number generation.
                  Generated automatically if not provided.

            options : Extra options. Accepts the following keyword arguments:

                * `epsilon` : Increment to `h0` to use for determining the
                  function gradient for `h0` update algorithm. The default
                  value is 1.4901161193847656e-08.
                * `maxiter` : Maximum number of iterations in the line search at the
                  optimal kernel bandwidth update. The default value is 10.
                * `momentum` : Momentum used in the next error calculation. Helps to
                  smooth out the change of error. The default value is 0.0.
                * `update_translations` : Update sample pixel translations
                  if True. The default value is False.
                * `return_extra` : Return errors and `h0` array if True. The default
                  value is False.

            verbose : Set verbosity of the computation process.

        Returns:
            A tuple of two items ('st_obj', 'extra'). The elements of the tuple
            are as follows:

            * `st_obj` : A new :class:`SpeckleTracking` object with the
              updated `pixel_map` and `reference_image`. `di_pix` and `dj_pix`
              are also updated if `update_translations` is True.

            * `extra`: A dictionary with the given parameters:

              * `errors` : List of average error values for each iteration.
              * `hvals` : List of kernel bandwidths for each iteration.

              Only if `return_extra` is True.
        """
        integrate = pm_args.get('integrate', False)

        epsilon = options.get('epsilon', 1e-3)
        maxiter = options.get('maxiter', 10)
        momentum = options.get('momentum', 0.0)
        update_translations = options.get('update_translations', False)
        return_extra = options.get('return_extra', False)

        obj = self.update_reference(hval=h0, method=ref_method)
        obj.update_errors.inplace_update()

        optimizer = BFGS(lambda hval: obj.CV(hval, ref_method),
                         np.atleast_1d(h0), epsilon=epsilon)

        errors = [obj.error,]
        hvals = [h0,]

        itor = tqdm(range(1, n_iter + 1), disable=not verbose,
                    bar_format='{desc} {percentage:3.0f}% {bar} ' \
                    'Iteration {n_fmt} / {total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]')
        if verbose:
            print(f"Initial error = {errors[-1]:.6f}, Initial h0 = {hvals[-1]:.2f}")

        for _ in itor:
            # Update pixel_map
            new_obj = obj.update_pixel_map(search_window=search_window, blur=blur,
                                           integrate=integrate, method=pm_method,
                                           extra_args=pm_args)

            # Update di_pix, dj_pix
            if update_translations:
                new_obj.update_translations.inplace_update(sw_y=search_window[0],
                                                           sw_x=search_window[1],
                                                           blur=blur)

            # Update hval and step
            optimizer.update_loss(lambda hval: new_obj.CV(hval, ref_method))
            optimizer.step(maxiter=maxiter)
            h0 = (1.0 - momentum) * optimizer.state_dict()['xk'].item() + momentum * hvals[-1]
            hvals.append(h0)

            # Update reference_image
            new_obj.update_reference.inplace_update(hval=h0, method=ref_method)

            new_obj.update_errors.inplace_update()
            errors.append((1.0 - momentum) * new_obj.error + momentum * errors[-1])
            if verbose:
                itor.set_description(f"Error = {errors[-1]:.6f}, hval = {hvals[-1]:.2f}")

            # Break if function tolerance is satisfied
            if (errors[-2] - errors[-1]) / max(errors[-2], errors[-1]) > f_tol:
                obj = new_obj

            else:
                break

        if return_extra:
            return obj, {'errors': errors, 'hvals': hvals}
        return obj

    def train(self, search_window: Tuple[float, float, float], h0: float,
              blur: float=0.0, n_iter: int=30, f_tol: float=1e-8,
              ref_method: str='KerReg', pm_method: str='rsearch',
              pm_args: Dict[str, Union[bool, int, float, str]]={},
              options: Dict[str, Union[bool, float, str]]={},
              verbose: bool=True) -> Tuple[SpeckleTracking, List[float]]:
        """Perform iterative Robust Speckle Tracking update. The reconstruction cycle
        consists of: (i) generating the reference image (see
        :func:`SpeckleTracking.update_reference`); (ii) updating the pixel mapping between
        a stack of frames and the generated reference image (see
        :func:`SpeckleTracking.update_pixel_map`); (iii) updating the sample translation
        vectors (if needed, see :func:`SpeckleTracking.update_translations`); and (iv)
        calculating the mean Huber error (see :func:`SpeckleTracking.update_error`). The
        kernel bandwidth in the reference update is kept fixed during the iterative update
        procedure.

        Args:
            search_window : A tuple of three elements ('sw_y', 'sw_x', 'sw_s'). The elements
                are the following:

                * `sw_y` : Search window size in pixels along the horizontal detector
                  axis.
                * `sw_x` : Search window size in pixels along the vertical detector
                  axis.
                * `sw_s` : Search window size of the Huber scaling map. Given as a ratio
                  (0.0 - 1.0) relative to the scaling map value before the update.

            h0 : Smoothing kernel bandwidth used in `reference_image` regression.
                The value is given in pixels.
            blur : Smoothing kernel bandwidth used in `pixel_map` regularisation.
                The value is given in pixels.
            n_iter : Maximum number of iterations.
            f_tol : Tolerance for termination by the change of the average error. The
                iteration stops when ``(f^k - f^{k+1})/max{|f^k|,|f^{k+1}|,1} <= ftol``.
            ref_method : `reference_image` update algorithm. The following keyword
                values are allowed:

                * `KerReg` : Kernel regression algorithm.
                * `LOWESS` : Local weighted linear regression.

            pm_method : `pixel_map` update algorithm. The following keyword
                values are allowed:

                * `gsearch` : Grid search algorithm.
                * `rsearch` : Random search algorithm.
                * `de`      : Differential evolution algorithm.

            pm_args : Pixel map update options. Accepts the following keyword
                arguments:

                * `integrate` : Ensure that the updated pixel map is irrotational
                  by integrating and taking the derivative. False by default.
                * `grid_size` : Grid size along one of the detector axes for
                  the 'gsearch' method. The grid shape is then (grid_size,
                  grid_size). The default value is :code:`int(0.5 * (sw_x + sw_y))`.
                * `n_trials` : Number of points generated at each  pixel in
                  the detector grid for the 'rsearch' method. The default value
                  is :code:`int(0.25 * (sw_x + sw_y)**2)`.
                * `n_iter` : The maximum number of generations over which the
                  entire population is evolved in the 'de' method. The default
                  value is 5.
                * `pop_size` : The total population size in the 'de' method.
                  Must be greater or equal to 4. The default value is
                  :code:`max(int(0.25 * (sw_x + sw_y)**2) / n_iter, 4)`.
                * `mutation` : The mutation constant in the 'de' method. It should
                  be in the range [0, 2]. The default value is 0.75.
                * `recombination` : The recombination constant  in the 'de' method,
                  should be in the range [0, 1]. The default value is 0.7.
                * `seed` : Specify seed for the random number generation.
                  Generated automatically if not provided.

            options : Extra options. Accepts the following keyword arguments:

                * `momentum` : Momentum used in the next error calculation. Helps to
                  smooth out the change of error.
                * `update_translations` : Update sample pixel translations
                  if True. The default value is False.
                * `return_extra` : Return errors at each iteration if True.
                  The default value is False.

            verbose : bool, optional
                Set verbosity of the computation process.

        Returns:
            A tuple of two items (`st_obj`, `errors`). The elements of the tuple
            are as follows:

            * `st_obj` : A new :class:`SpeckleTracking` object with the updated
              `pixel_map` and `reference_image`. `di_pix` and `dj_pix`
              are also updated if 'update_translations' in `options`
              is True.

            * `errors` : List of average error values for each iteration. Only if
              'return_extra' in `options` is True.
        """
        integrate = pm_args.get('integrate', False)

        momentum = options.get('momentum', 0.0)
        update_translations = options.get('update_translations', False)
        return_extra = options.get('return_extra', False)

        obj = self.update_reference(hval=h0, method=ref_method)
        obj.update_errors.inplace_update()
        errors = [obj.error]

        itor = tqdm(range(1, n_iter + 1), disable=not verbose,
                    bar_format='{desc} {percentage:3.0f}% {bar} '\
                    'Iteration {n_fmt} / {total_fmt} [{elapsed}<{remaining}, {rate_fmt}{postfix}]')

        if verbose:
            print(f"Initial error = {errors[0]:.6f}, "\
                  f"Initial h0 = {h0:.2f}")

        for _ in itor:
            # Update pixel_map
            new_obj = obj.update_pixel_map(search_window=search_window, blur=blur,
                                           integrate=integrate, method=pm_method,
                                           extra_args=pm_args)

            # Update di_pix, dj_pix
            if update_translations:
                new_obj.update_translations.inplace_update(sw_y=search_window[0],
                                                           sw_x=search_window[1],
                                                           blur=blur)

            # Update reference_image
            new_obj.update_reference.inplace_update(hval=h0, method=ref_method)
            new_obj.update_errors.inplace_update()

            # Calculate errors
            errors.append((1.0 - momentum) * new_obj.error + momentum * errors[-1])
            if verbose:
                itor.set_description(f"Error = {errors[-1]:.6f}")

            # Break if function tolerance is satisfied
            if (errors[-2] - errors[-1]) / max(errors[-2], errors[-1]) > f_tol:
                obj = new_obj

            else:
                break

        if return_extra:
            return obj, errors
        return obj

    def CV(self, hval: float, method: str='KerReg') -> float:
        """Return cross-validation error for the given kernel bandwidth `hval`.
        The cross-validation error metric is calculated as follows: (i) generate
        a reference profile based on the training subset and (ii) find the
        mean-squared-error (MSE) for the test subset.

        Args:
            hval : `reference_image` kernel bandwidths in pixels.
            method : `reference_image` update algorithm. The following keyword
                values are allowed:

                * `KerReg` : Kernel regression algorithm.
                * `LOWESS` : Local weighted linear regression.

        Returns:
            Cross-validation error.

        Raises:
            ValueError : If `method` keyword value is not valid.

        See Also:
            :func:`pyrost.bin.pm_total_error` : Full details of the error metric.
        """
        if method == 'KerReg':
            I0, n0, m0 = KR_reference(I_n=self.data, W=self.whitefield * self.train_mask,
                                      u=self.pixel_map, di=self.di_pix, dj=self.dj_pix,
                                      ds_y=self.ds_y, ds_x=self.ds_x, hval=hval,
                                      num_threads=self.num_threads)
        elif method == 'LOWESS':
            I0, n0, m0 = LOWESS_reference(I_n=self.data, W=self.whitefield * self.train_mask,
                                          u=self.pixel_map, di=self.di_pix, dj=self.dj_pix,
                                          ds_y=self.ds_y, ds_x=self.ds_x, hval=hval,
                                          num_threads=self.num_threads)
        else:
            raise ValueError('Method keyword is invalid')
        error = pm_total_error(I_n=self.data, W=self.whitefield * self.test_mask, I0=I0,
                               u=self.pixel_map, di=self.di_pix - n0, dj=self.dj_pix - m0,
                               ds_y=self.ds_y, ds_x=self.ds_x, sigma=self.scale_map,
                               num_threads=self.num_threads)
        return error

    def CV_curve(self, harr: np.ndarray, method: str='KerReg', verbose: bool=True) -> np.ndarray:
        """Return a set of cross-validation errors for a set of kernel
        bandwidths.

        Args:
            harr : Set of `reference_image` kernel bandwidths in pixels.
            method : `reference_image` update algorithm. The following keyword
                values are allowed:

                * `KerReg` : Kernel regression algorithm.
                * `LOWESS` : Local weighted linear regression.

            verbose : Set the verbosity of the process.

        Returns:
            An array of cross-validation errors.

        See Also:
            :func:`pyrost.bin.pm_total_error` : Full details of the error metric.
        """
        mse_list = []
        for hval in tqdm(np.array(harr, ndmin=1), total=len(harr),
                         disable=not verbose, desc='Calculating CV'):
            mse_list.append(self.CV(hval, method=method))
        return np.array(mse_list)




"""Object oriented implementation of the Broyden-Fletcher-Goldfarb-Shanno (BFGS) algorithm
for a function of one parameter.
"""
from typing import Callable, Dict, Optional, Tuple, Union
from copy import deepcopy
from scipy.optimize import approx_fprime
from scipy.optimize.minpack2 import dcsrch
from numpy.linalg import norm
import numpy as np

class BFGS():
    """Minimize a function of one parameter using the the Broyden-Fletcher-
    Goldfarb-Shanno (BFGS) algorithm.
    """
    params: Dict[str, Union[int, float, np.ndarray]]

    def __init__(self, loss: Callable[[np.ndarray], float], x0: np.ndarray,
                 grad: Optional[Callable[[np.ndarray], float]]=None,
                 epsilon=1e-4, c1: float=1e-4, c2: float=0.9, xtol=1e-14,
                 line_search: str='minpack'):
        """
        Args:
            loss : Objective function to be minimized.
            x0 : Initial guess.
            grad : Gradient of the objective function.
            epsilon : If `grad` is approximated, use this value for the step
                size.
            c1 : Parameter for Armijo condition rule.
            c2 : Parameter for curvature condition rule.
            xtol : Relative tolerance for an acceptable step in the line search
                algorithm.
            line_search: Choose the implementation of the line search algorithm 
                to enforce strong Wolfe conditions. The following keyword values
                are allowed:

                * `minpack` : MINPACK line search algorithm.
                * `scipy` : SciPy line search algorithm.
        """
        if line_search == 'minpack':
            self.line_search = self._line_search_minpack
        elif line_search == 'scipy':
            self.line_search = self._line_search_scipy
        else:
            raise ValueError('Invalid line_search keyword argument')

        self._p = {'fcount': 0, 'gcount': 0, 'xk': x0,
                   'c1': c1, 'c2': c2, 'xtol': xtol, 'epsilon': epsilon}

        self.update_loss(loss, grad)

        self._p['fval'] = self.loss(self._p['xk'])
        self._p['I'] = np.eye(len(self._p['xk']), dtype=int)
        self._p['Hk'] = np.eye(len(self._p['xk']))
        self._update_gnorm()
        self._p['old_fval'] = self._p['fval'] + 0.5 * norm(self._p['gfk'])

    def update_loss(self, loss: Callable[[np.ndarray], float],
                    grad: Optional[Callable[[np.ndarray], float]]=None) -> None:
        """Update the objective function to minimize.

        Args:
            loss : Objective function to be minimized.
            grad : Gradient of the objective function.
        """
        if grad is None:
            grad = lambda x: approx_fprime(x, loss, self._p['epsilon'])
        self._loss = loss
        self._grad = grad
        self._p['gfk'] = self.grad(self._p['xk'])
        self._p['gfkp1'] = self._p['gfk']

    def loss(self, x: np.ndarray) -> float:
        """Return the objective value for a given argument.

        Args:
            x : Argument value.

        Returns:
            Objective value
        """
        self._p['fcount'] += 1
        return self._loss(x)

    def grad(self, x: np.ndarray) -> float:
        """Return the gradient value of the objective function for a given
        argument.

        Args:
            x : Argument value.

        Returns:
            A gradient value of the objective function.
        """
        self._p['gcount'] += 1
        return self._grad(x)

    def _update_pk(self):
        self._p['pk'] = -np.dot(self._p['Hk'], self._p['gfk'])

    def _update_gnorm(self):
        self._p['gnorm'] = np.amax(np.abs(self._p['gfk']))

    def _phi(self, s: np.ndarray) -> float:
        return self.loss(self._p['xk'] + s * self._p['pk'])

    def _derphi(self, s: np.ndarray) -> float:
        self._p['gfkp1'] = self.grad(self._p['xk'] + s * self._p['pk'])
        return np.dot(self._p['gfkp1'], self._p['pk'])

    @staticmethod
    def _cubicmin(a: float, fa: float, fpa: float, b: float, fb: float, c: float, fc: float) -> Union[float, None]:
        with np.errstate(divide='raise', over='raise', invalid='raise'):
            try:
                db = b - a
                dc = c - a
                denom = (db * dc) ** 2 * (db - dc)
                d1 = np.empty((2, 2))
                d1[0, 0] = dc ** 2
                d1[0, 1] = -db ** 2
                d1[1, 0] = -dc ** 3
                d1[1, 1] = db ** 3
                [A, B] = np.dot(d1, np.asarray([fb - fa - fpa * db,
                                                fc - fa - fpa * dc]).flatten())
                A /= denom
                B /= denom
                radical = B * B - 3.0 * A * fpa
                xmin = a + (-B + np.sqrt(radical)) / (3.0 * A)
            except ArithmeticError:
                return None
        if not np.isfinite(xmin):
            return None
        return xmin

    @staticmethod
    def _quadmin(a: float, fa: float, fpa: float, b: float, fb: float) -> Union[float, None]:
        with np.errstate(divide='raise', over='raise', invalid='raise'):
            try:
                db = b - a
                xmin = a - fpa / (2.0 * (fb - fa - fpa * db) / (db * db))
            except ArithmeticError:
                return None
        if not np.isfinite(xmin):
            return None
        return xmin

    def _zoom(self, a_lo: float, a_hi: float, phi_lo: float, phi_hi: float, derphi_lo: float,
              phi0: float, derphi0: float, maxiter: int=10) -> Tuple[float, float]:
        phi_rec = phi0
        a_rec = 0.0
        for i in range(maxiter):
            # interpolate to find a trial step length between a_lo and
            # a_hi Need to choose interpolation here. Use cubic
            # interpolation and then if the result is within delta *
            # dalpha or outside of the interval bounded by a_lo or a_hi
            # then use quadratic interpolation, if the result is still too
            # close, then use bisection

            dalpha = a_hi - a_lo
            if dalpha < 0.0:
                a, b = a_hi, a_lo
            else:
                a, b = a_lo, a_hi

            # minimizer of cubic interpolant
            # (uses phi_lo, derphi_lo, phi_hi, and the most recent value of phi)
            #
            # if the result is too close to the end points (or out of the
            # interval), then use quadratic interpolation with phi_lo,
            # derphi_lo and phi_hi if the result is still too close to the
            # end points (or out of the interval) then use bisection

            if i > 0:
                cchk = 0.2 * dalpha # cubic interpolant check
                a_j = self._cubicmin(a_lo, phi_lo, derphi_lo, a_hi, phi_hi,
                                     a_rec, phi_rec)
            if i == 0 or a_j is None or a_j > b - cchk or a_j < a + cchk:
                qchk = 0.1 * dalpha # quadratic interpolant check
                a_j = self._quadmin(a_lo, phi_lo, derphi_lo, a_hi, phi_hi)
                if a_j is None or a_j > b - qchk or a_j < a + qchk:
                    a_j = a_lo + 0.5 * dalpha

            # Check new value of a_j

            phi_aj = self._phi(a_j)
            if (phi_aj > phi0 + self._p['c1'] * a_j * derphi0) or (phi_aj >= phi_lo):
                phi_rec = phi_hi
                a_rec = a_hi
                a_hi = a_j
                phi_hi = phi_aj
            else:
                derphi_aj = self._derphi(a_j)
                if abs(derphi_aj) <= -self._p['c2'] * derphi0:
                    break
                if derphi_aj * (a_hi - a_lo) >= 0.0:
                    phi_rec = phi_hi
                    a_rec = a_hi
                    a_hi = a_lo
                    phi_hi = phi_lo
                else:
                    phi_rec = phi_lo
                    a_rec = a_lo
                a_lo = a_j
                phi_lo = phi_aj
                derphi_lo = derphi_aj

        a_star = a_j
        val_star = phi_aj

        return a_star, val_star

    def _line_search_scipy(self, maxiter: int, amin: float, amax: float):
        phi0 = self._p['fval']
        old_phi0 = self._p['old_fval']
        derphi0 = self._derphi(0.0)

        if derphi0 != 0.0:
            alpha1 = min(1.0, 2.02 * (phi0 - old_phi0) / derphi0)
        else:
            alpha1 = 1.0

        if alpha1 <= amin:
            alpha1 = 1.0

        alpha0 = 0.0
        alpha1 = min(alpha1, amax)
        phi_a1 = self._phi(alpha1)
        phi_a0 = phi0
        derphi_a0 = derphi0

        for i in range(maxiter):
            if alpha1 == amin or alpha0 == amax:
                break

            if phi_a1 > phi0 + self._p['c1'] * alpha1 * derphi0 or (phi_a1 >= phi_a0 and i > 0):
                alpha_star, phi_star = self._zoom(alpha0, alpha1, phi_a0,
                                                  phi_a1, derphi_a0, phi0, derphi0)
                break

            derphi_a1 = self._derphi(alpha1)
            if abs(derphi_a1) <= -self._p['c2'] * derphi0:
                alpha_star = alpha1
                phi_star = phi_a1
                break

            if derphi_a1 >= 0.0:
                alpha_star, phi_star = self._zoom(alpha1, alpha0, phi_a1,
                                                  phi_a0, derphi_a1, phi0, derphi0)
                break

            # increase by factor of two on each iteration
            alpha0 = max(alpha1, amin)
            alpha1 = min(2.0 * alpha1, amax)
            phi_a0 = phi_a1
            phi_a1 = self._phi(alpha1)
            derphi_a0 = derphi_a1
        else:
            alpha_star = alpha1
            phi_star = phi_a1

        # stopping test maxiter reached
        self._p['alpha_k'] = alpha_star
        self._p['fval'] = phi_star
        self._p['old_fval'] = phi0

    def _line_search_minpack(self, maxiter: int, amin: float, amax: float):
        phi0 = self._p['fval']
        old_phi0 = self._p['old_fval']
        derphi0 = self._derphi(0.0)

        if derphi0 != 0.0:
            alpha1 = min(1.0, 2.02 * (phi0 - old_phi0) / derphi0)
        else:
            alpha1 = 1.0

        if alpha1 <= amin:
            alpha1 = 1.0

        phi1 = phi0
        derphi1 = derphi0
        isave = np.zeros((2,), np.intc)
        dsave = np.zeros((13,), float)
        task = b'START'

        for _ in range(maxiter):
            stp, phi1, derphi1, task = dcsrch(alpha1, phi1, derphi1, self._p['c1'],
                                              self._p['c2'], self._p['xtol'],
                                              task, amin, amax, isave, dsave)

            if task[:2] == b'FG':
                alpha1 = stp
                phi1 = self._phi(stp)
                derphi1 = self._derphi(stp)
            else:
                break

        self._p['alpha_k'] = stp
        self._p['fval'] = phi1
        self._p['old_fval'] = phi0

    def step(self, maxiter: int=10, amin: float=1e-100, amax: float=1e100):
        """Performs a single optimization step.

        Args:
            maxiter : Maximum number of iteration of the line search algorithm
                to perform.
            amin : Minimum step size.
            amax : Maximum step size.
        """
        self._update_pk()
        self.line_search(maxiter=maxiter, amin=amin, amax=amax)

        xkp1 = self._p['xk'] + self._p['alpha_k'] * self._p['pk']
        sk = xkp1 - self._p['xk']
        self._p['xk'] = xkp1

        yk = self._p['gfkp1'] - self._p['gfk']
        self._p['gfk'] = self._p['gfkp1']
        self._update_gnorm()

        rhok_inv = np.dot(yk, sk)
        if rhok_inv == 0.0:
            rhok = 1000.0
        else:
            rhok = 1.0 / rhok_inv

        A1 = self._p['I'] - sk[:, np.newaxis] * yk[np.newaxis, :] * rhok
        A2 = self._p['I'] - yk[:, np.newaxis] * sk[np.newaxis, :] * rhok
        new_Hk = np.dot(A1, np.dot(self._p['Hk'], A2)) + \
                        (rhok * sk[:, np.newaxis] * sk[np.newaxis, :])

        if np.trace(self._p['Hk']) > 0.0:
            self._p['Hk'] = new_Hk

    def state_dict(self):
        """Returns the state of the optimizer as a dict.

        Returns:
            A dictionary with all the parameters of the optimizer:

            * `fcount` : Number of functions evaluations made.
            * `gcount` : Number of gradient evaluations made.
            * `c1` : Parameter for Armijo condition rule.
            * `c2` : Parameter for curvature condition rule.
            * `xk` : The current point.
            * `fval` : Objective value of the current point.
            * `old_fval` : Objective value of the point prior to 'xk'.
            * `gfk` : Gradient value of the current point.
            * `gnorm` : Gradient norm value of the current point.
            * `Hk` : The current guess of the Hessian matrix.
            * `epsilon` : If `grad` is approximated, use this value for the
                step size.
            * `xtol` :  Relative tolerance for an acceptable step in the line
                search algorithm.
        """
        return deepcopy(self._p)



def pytest_addoption(parser):
    parser.addoption("--scan_num", type=int, default=2989,
                     help="Scan number")
    parser.addoption("--roi", type=int, nargs=4, default=(270, 300, 200, 1240),
                     help="Region of interest")
    parser.addoption("--good_frames", type=int, nargs=2, default=(5, 100),
                     help="Range of good frames")
    parser.addoption("--defocus", type=float, default=152e-6,
                     help="Defocus distance [m]")
    parser.addoption("--distance", type=float, default=2.0,
                     help="Defocus distance [m]")
    parser.addoption("--wavelength", type=float, default=7.092917530503447e-11,
                     help="X-ray beam wavelength [m]")

def pytest_generate_tests(metafunc):
    # This is called for every test. Only get/set command line arguments
    # if the argument is specified in the list of test "fixturenames".
    option_dict = vars(metafunc.config.option)
    for attr in ('scan_num', 'roi', 'good_frames', 'defocus', 'distance', 'wavelength'):
        option_value = option_dict.get(attr)
        if attr in metafunc.fixturenames and option_value is not None:
            metafunc.parametrize(attr, [option_value])
